
local rematch = Rematch
local settings

-- the unsorted leveling queue is settings.LevelingQueue
rematch.levelingPetsByID = {} -- for faster lookups, this is all pets in the queue indexed by petID
-- which is copied to sortedQueue and sorted if need be
rematch.sortedQueue = {}

rematch.levelingIcon = "Interface\\AddOns\\Rematch\\leveling"

function rematch:InitializeLeveling()
	settings = RematchSettings

	rematch.drawer.leveling.set = 3 -- set 3 is leveling pet
	rematch.drawer.leveling:RegisterForClicks("LeftButtonUp","RightButtonUp")
	rematch.drawer.leveling:RegisterForDrag("LeftButton")

	rematch.drawer.leveling.leveling:Show() -- SetShown(petID~=nil)

	-- texcoords for sort control buttons and tooltip info
	local sortControlInfo = {
		{REMATCH_TOOLTIP_SORTASCEND_TITLE,REMATCH_TOOLTIP_SORTASCEND_TEXT,0.57812500,0.82812500,0.41406250,0.28906250},
		{REMATCH_TOOLTIP_SORTDESCEND_TITLE,REMATCH_TOOLTIP_SORTDESCEND_TEXT,0.57812500,0.82812500,0.28906250,0.41406250},
		{REMATCH_TOOLTIP_SORTMEDIAN_TITLE,REMATCH_TOOLTIP_SORTMEDIAN_TEXT,0.29687500,0.54687500,0.28906250,0.41406250},
		{REMATCH_TOOLTIP_AUTOROTATE_TITLE,REMATCH_TOOLTIP_AUTOROTATE_TEXT,0.01562500,0.26562500,0.13281250,0.00781250},
	}
	for i=1,4 do
		local button = rematch.queue.controls.sortButtons[i]
		button.icon:SetTexture("Interface\\Common\\UI-ModelControlPanel")
		button.icon:SetTexCoord(select(3,unpack(sortControlInfo[i])))
		button.tooltipTitle = sortControlInfo[i][1]
		button.tooltipBody = sortControlInfo[i][2]
	end

	-- setup list scrollFrame
	local scrollFrame = rematch.queue.scrollFrame
	scrollFrame.update = rematch.UpdateQueueList
	scrollFrame.stepSize = 180
	HybridScrollFrame_CreateButtons(scrollFrame, "RematchQueueListTemplate")
	local scrollBar = scrollFrame.scrollBar
	scrollBar.doNotHide = true
	scrollBar.trackBG:SetPoint("TOPLEFT",-2,25)
	scrollBar.trackBG:SetPoint("BOTTOMRIGHT",2,-25)
	scrollBar.trackBG:SetTexture("Interface\\ChatFrame\\ChatFrameBackground")
	scrollBar.trackBG:SetGradientAlpha("HORIZONTAL",.125,.125,.125,1,.05,.05,.05,1)

	scrollFrame.insertLine.timer = 0

	rematch.queue:SetFrameLevel(rematch.queue:GetFrameLevel()+6)
	rematch.queue.emptyCapture:SetFrameLevel(scrollFrame:GetFrameLevel()+2)
	scrollFrame.insertLine:SetFrameLevel(scrollFrame:GetFrameLevel()+3)
	rematch.queue.resultsBar:SetFrameLevel(scrollFrame:GetFrameLevel()+4)

	rematch:ColorControlButton(rematch.queue.closeButton,.85)
	rematch.queue.emptyCapture:RegisterForClicks("AnyUp")
	rematch:ProcessQueue()

	rematch.queue.prevPet.tooltipTitle = REMATCH_TOOLTIP_PREVPET
	rematch.queue.nextPet.tooltipTitle = REMATCH_TOOLTIP_NEXTPET
	rematch.levelingSwitcher.prevPet.tooltipTitle = REMATCH_TOOLTIP_PREVPET
	rematch.levelingSwitcher.nextPet.tooltipTitle = REMATCH_TOOLTIP_NEXTPET
	rematch.queue.controls.fillButton.tooltipTitle = REMATCH_TOOLTIP_FILL_TITLE
	rematch.queue.controls.fillButton.tooltipBody = REMATCH_TOOLTIP_FILL_TEXT
	rematch.queue.controls.emptyButton.tooltipTitle = REMATCH_TOOLTIP_EMPTY_TITLE
	rematch.queue.controls.emptyButton.tooltipBody = REMATCH_TOOLTIP_EMPTY_TEXT

	rematch:SetCurrentLevelingPet(settings.CurrentLevelingPet)
end

function rematch:RematchQueueToggleButton()
	local showQueue = not rematch.queue:IsVisible()
	rematch.queue:SetShown(showQueue)
	settings.queueOpen = showQueue
	rematch:PositionQueueForSidePanel()
end

-- also called in UpdateBrowserPosition
function rematch:PositionQueueForSidePanel()
	local sidePanel = rematch.sidePanel
	if settings.sidePanelOpen and settings.queueOpen and sidePanel:IsVisible() then
		rematch.queue:ClearAllPoints()
		if rematch.sidePanel:GetPoint()=="TOPRIGHT" then
			rematch.queue:SetPoint("BOTTOMRIGHT",rematch.sidePanel,"BOTTOMLEFT",2,0)
		else
			rematch.queue:SetPoint("BOTTOMLEFT",rematch.sidePanel,"BOTTOMRIGHT",-2,0)
		end
	end
end

--[[ Leveling Pet System ]]

-- returns true if the petID is in the queue
function rematch:IsPetLeveling(petID)
	return petID and rematch.levelingPetsByID[petID]
end

-- returns true if the petID is the *current* leveling pet
function rematch:IsCurrentLevelingPet(petID)
	return petID and settings.CurrentLevelingPet==petID
end

-- returns the petID of the current leveling pet
function rematch:GetCurrentLevelingPet()
	return settings.CurrentLevelingPet
end

-- returns true if petID is below 25 and can battle
-- also returns pet's level (including partial level)
function rematch:PetCanLevel(petID)
	if petID and type(petID)=="string" then
		local _,_,level,xp,maxXp,_,_,_,_,_,_,_,_,_,canBattle = C_PetJournal.GetPetInfoByPetID(petID)
		if level and level<25 and canBattle then
			return true,level+(xp/maxXp)
		end
	end
end

-- returns either the icon of the pet, or "interface\\addons\\rematch\\leveling.tga" if no leveling pet
function rematch:GetLevelingPetIcon()
	local petID = settings.CurrentLevelingPet
	local icon = petID and (select(9,C_PetJournal.GetPetInfoByPetID(petID))) or rematch.levelingIcon
	return icon
end

function rematch:GetNumLevelingPets()
	return #rematch.sortedQueue
end

-- returns the index in sortedQueue for the current leveling pet
function rematch:GetCurrentQueuePosition()
	local sorted = rematch.sortedQueue
	local petID = settings.CurrentLevelingPet
	for i=1,#sorted do
		if sorted[i]==petID then
			return i
		end
	end
end

--[[ Leveling Slot ]]

function rematch:UpdateLevelingSlot()
	local currentPetID = rematch:GetCurrentLevelingPet()
	rematch.drawer.leveling.petID = currentPetID
	rematch.drawer.leveling.icon:SetTexture(rematch:GetLevelingPetIcon())
	if currentPetID then
		rematch.drawer.leveling.level.text:SetText((select(3,C_PetJournal.GetPetInfoByPetID(currentPetID))))
	end
	rematch.drawer.leveling.level:SetShown(currentPetID and true)
end

function rematch:LevelingSlotOnEnter()
	local cursorType,petID = GetCursorInfo()
	local canLevel = cursorType=="battlepet" and rematch:PetCanLevel(petID)
	if not rematch:GetCurrentLevelingPet() then
		if canLevel then
			return -- don't show helpplate (or floatingpetcard) if a pet that can level is over empty slot
		end
		HelpPlate_TooltipHide()
		HelpPlateTooltip.ArrowLEFT:Show()
		HelpPlateTooltip.ArrowGlowLEFT:Show()
		HelpPlateTooltip:SetPoint("RIGHT", self, "LEFT", -10, 0)
		HelpPlateTooltip.Text:SetText(Rematch.helpPlate[5].ToolTipText)
		HelpPlateTooltip:Show()
	else
		if not canLevel and cursorType=="battlepet" then
			rematch:ShowTooltip(REMATCH_QUEUE_WARNING_CANTLEVEL_TITLE,nil,true)
		end
		rematch:ShowFloatingPetCard(self.petID,self)
	end
end

function rematch:LevelingSlotOnLeave()
	HelpPlate_TooltipHide()
	rematch.PetOnLeave(self)
	RematchTooltip:Hide()
end

function rematch:LevelingSlotOnClick(button)
	if button=="RightButton" and GetCursorInfo()=="battlepet" then
		ClearCursor()
	elseif button=="RightButton" and rematch:GetCurrentLevelingPet() then
		rematch.dropDownPetID = self.petID
		rematch.dropDownPetSlot = nil
		rematch:ShowDropDown(3)
	else
		rematch.PetOnClick(self,button)
	end
end

function rematch:UpdateLevelingBorders()
	local anyleveling
	local switcher = rematch.levelingSwitcherMouseover
	for i=1,3 do
		local current = rematch.current.pet[i]
		local petID = current.petID
		if rematch:IsCurrentLevelingPet(petID) then
			current.leveling:Show()
			if switcher:GetParent()~=current then
				switcher:SetParent(current)
				switcher:SetPoint("TOP",current,"BOTTOM",0,1)
			end
			anyleveling = true
		else
			current.leveling:Hide()
		end
	end
end

--[[ Leveling Queue Frame ]]

function rematch:QueueOnShow()
	rematch:ProcessQueue()
	rematch:UpdateQueueControls()
	rematch:ScrollToQueueIndex(rematch:GetCurrentQueuePosition())
end

function rematch:UpdateQueueControls()
	rematch.queue.controls:SetShown(settings.UseQueueControls)
	rematch.queue.scrollFrame:SetHeight(settings.UseQueueControls and 294 or 344)
	local order = settings.QueueSortOrder
	for i=1,3 do
		local button = rematch.queue.controls.sortButtons[i]
		button:SetChecked(order==i)
		if order then
			rematch:DesaturateControlButton(button,order~=i)
		else
			rematch:DesaturateControlButton(button,false)
		end
	end
	rematch:UpdateQueueList()
	if settings.UseQueueControls then
		rematch.queue.toggle:SetNormalTexture("Interface\\ChatFrame\\UI-ChatIcon-ScrollUp-Up")
		rematch.queue.toggle:SetPushedTexture("Interface\\ChatFrame\\UI-ChatIcon-ScrollUp-Down")
	else
		rematch.queue.toggle:SetNormalTexture("Interface\\ChatFrame\\UI-ChatIcon-ScrollDown-Up")
		rematch.queue.toggle:SetPushedTexture("Interface\\ChatFrame\\UI-ChatIcon-ScrollDown-Down")
	end
	rematch.queue.controls.sortButtons[4]:SetChecked(settings.QueueAutoRotate)
end

function rematch:ToggleQueueControls()
	settings.UseQueueControls = not settings.UseQueueControls
	rematch:UpdateQueueControls()
end

-- when pet dragged over QueueListButton (or capture area)
function rematch:QueueListOnEnter()
	if self and self.petID then
		rematch:ShowFloatingPetCard(self.petID,self)
	end
	local cursorType,petID = GetCursorInfo()
	if cursorType=="battlepet" then
		if settings.QueueSortOrder and rematch:IsPetLeveling(petID) then
			rematch:ShowTooltip(REMATCH_QUEUE_WARNING_SORTED_TITLE,REMATCH_QUEUE_WARNING_SORTED_TEXT,true)
			return -- can't reorder leveling pets while queue has a sort order
		elseif rematch:PetCanLevel(petID) then
			if settings.QueueSortOrder then
				rematch:ShowTooltip(REMATCH_QUEUE_WARNING_OKBUTSORTED_TITLE,REMATCH_QUEUE_WARNING_OKBUTSORTED_TEXT,true)
			end
			rematch.queue.scrollFrame.insertLine.timer = 1 -- start immediately
			rematch.queue.scrollFrame.insertLine.parent = self
			rematch.queue.scrollFrame.insertLine:Show()
		else
			rematch:ShowTooltip(REMATCH_QUEUE_WARNING_CANTLEVEL_TITLE,nil,true)
		end
	end
end

function rematch:QueueListOnLeave()
	rematch:HideFloatingPetCard()
	rematch.queue.scrollFrame.insertLine:Hide()
	RematchTooltip:Hide()
end

-- when a pet within the queue is clicked
function rematch:QueueListOnClick(button)
	local cursorType = GetCursorInfo()
	if button=="RightButton" and cursorType~="battlepet" and self.petID then
		rematch.dropDownPetID = self.petID
		rematch:ShowDropDown(4)
	else
		if cursorType=="battlepet" then
			rematch.QueueOnReceiveDrag(self) -- pet being dropped
		else
			rematch.PetOnClick(self,button) -- do left-click stuff
		end
	end
end

function rematch:QueueListOnDoubleClick(button)
	rematch:StartLevelingPet(self.petID)
	rematch:HideFloatingPetCard(true)
end

-- updates the hybridscrollframe in the levelingqueue
function rematch:UpdateQueueList()
	local queue = rematch.sortedQueue
	local numData = #queue
	local scrollFrame = rematch.queue.scrollFrame
	local offset = HybridScrollFrame_GetOffset(scrollFrame)
	local buttons = scrollFrame.buttons
	local lastButton
	local cursorType,petOnCursor = GetCursorInfo()
	petOnCursor = cursorType=="battlepet" and petOnCursor

	for i=1, #buttons do
		local index = i + offset
		local button = buttons[i]
		if ( index <= numData) then
			button.petID = queue[index]
			button.index = index
			local petID = button.petID
			local onCursor = petOnCursor==petID
			local _,_,level,_,_,_,_,_,icon,petType = C_PetJournal.GetPetInfoByPetID(petID)
			if icon then
				local isSelected = button.petID==rematch:GetCurrentLevelingPet()
				button.icon:SetTexture(icon)
				button.level:SetText(level)
				local rarity = select(5,C_PetJournal.GetPetStats(petID))
		    local r,g,b = ITEM_QUALITY_COLORS[rarity-1].r, ITEM_QUALITY_COLORS[rarity-1].g, ITEM_QUALITY_COLORS[rarity-1].b
				if onCursor then
					button.rarity:SetGradientAlpha("VERTICAL",.4,.4,.4,.25,.4,.4,.4,.6)
				elseif isSelected then
					button.rarity:SetGradientAlpha("VERTICAL",1,.82,0,.5,r,g,b,.6)
				else
			    button.rarity:SetGradientAlpha("VERTICAL",r,g,b,.25,r,g,b,.6)
				end
				button.type:SetTexture("Interface\\PetBattles\\PetIcon-"..PET_TYPE_SUFFIX[petType])
				if button==GetMouseFocus() then
					rematch.QueueListOnEnter(button) -- we're scrolling, pretend moving into this button
				end
				button.icon:SetDesaturated(onCursor)
				button.type:SetDesaturated(onCursor)
			end
			button.selected:SetShown(button.petID and button.petID==rematch:GetCurrentLevelingPet())
			lastButton = button
			button:Show()
		else
			button.petID = nil
			button.index = nil
			button:Hide()
		end
	end

	-- a scrollframe can't receive clicks/drags; this invisible button will capture clicks/drags to empty area of scrollframe
	if not scrollFrame.scrollBar:IsEnabled() then
		if lastButton then
			rematch.queue.emptyCapture:SetPoint("TOPLEFT",lastButton,"BOTTOMLEFT",0,0)
		else
			rematch.queue.emptyCapture:SetPoint("TOPLEFT",scrollFrame,"TOPLEFT")
		end
		rematch.queue.emptyCapture:Show()
	else
		rematch.queue.emptyCapture:Hide()
	end

	rematch.queue.resultsBar.count:SetText(numData)

	HybridScrollFrame_Update(scrollFrame,30*numData,30)
end

function rematch:RematchQueueFillButton()
	if settings.AutoConfirmFill then
		rematch:FillQueue()
	else
		StaticPopupDialogs["REMATCH_FILL"]=StaticPopupDialogs["REMATCH_FILL"] or { text=REMATCH_FILLCONFIRM, button1=YES, button2=NO, hideOnEscape=1, timeout=0, whileDead=1, preferredIndex=3, OnAccept=rematch.FillQueue }
		local count = rematch:FillQueue(true)
		local extraText = count>100 and REMATCH_FILLCONFIRM_MANY or count==0 and REMATCH_FILLCONFIRM_NONE  or ""
		StaticPopup_Show("REMATCH_FILL",count,extraText)
	end
end

-- fill button on queuecontrols; adds one of each species listed in the browser
function rematch:FillQueue(countOnly)
	local roster = rematch.roster
	local queue = settings.LevelingQueue
	local scratch = rematch.info
	rematch:ProcessQueue()
	roster:Update()
	wipe(scratch)
	-- count how many species to add
	local count = 0
	for i=1,roster:GetNumPets() do
		local petID = roster:GetPetByIndex(i)
		local speciesID,level,canBattle,_
		if type(petID)=="string" then -- owned pet
			speciesID, _, level, _, _, _, _, _, _, _, _, _, _, _, canBattle = C_PetJournal.GetPetInfoByPetID(petID)
			if rematch:IsPetLeveling(petID) then
				scratch[speciesID]=1
			elseif rematch:PetCanLevel(petID) and not scratch[speciesID] then
				if not countOnly then
					table.insert(queue,petID)
				end
				count = count + 1
				scratch[speciesID]=1
			end
		end
	end
	rematch:ProcessQueue()
	rematch:UpdateQueueList()
	rematch:UpdateLevelingMarkers()
	if countOnly then
		return count
	end
	rematch.queue.resultsBar.shine:SetCooldown(0,1)
end

function rematch:RematchQueueEmptyButton()
	if settings.AutoConfirmFill then
		rematch:EmptyQueue()
	else
		StaticPopupDialogs["REMATCH_EMPTY"]=StaticPopupDialogs["REMATCH_EMPTY"] or { text=REMATCH_EMPTYCONFIRM, button1=YES, button2=NO, hideOnEscape=1, timeout=0, whileDead=1, preferredIndex=3, OnAccept=rematch.EmptyQueue }
		StaticPopup_Show("REMATCH_EMPTY")
	end
end

function rematch:EmptyQueue()
	wipe(settings.LevelingQueue)
	if not settings.KeepQueueSort then
		settings.QueueSortOrder = nil
		settings.QueueAutoRotate = nil
		rematch:UpdateQueueControls()
	end
	rematch:ProcessQueue()
	rematch:UpdateQueueList()
	rematch:UpdateLevelingMarkers()
	rematch.queue.resultsBar.shine:SetCooldown(0,1)
end

--[[ queue manipulation ]]

-- sets CurrentLevelingPet, swaps loadout if needed, updates slot
-- note: petID must already be in queue or it will do nothing!
-- use StartLevelingPet(petID) instead
function rematch:SetCurrentLevelingPet(petID)
	local oldLevelingPet = settings.CurrentLevelingPet
	if petID~=oldLevelingPet and rematch:IsPetLeveling(petID) then
		settings.CurrentLevelingPet = petID
		if petID then -- oldLevelingPet and
			local loadoutSlot
			for i=1,3 do
				local loadoutPetID = C_PetJournal.GetPetLoadOutInfo(i)
				if loadoutPetID==oldLevelingPet or rematch:IsPetLeveling(loadoutPetID) then
					loadoutSlot = i
					if loadoutPetID==oldLevelingPet then
						break -- this slot is old current leveling pet, ignore rest
					end
				end
			end
			if loadoutSlot then
				if not C_PetBattles.IsInBattle() and not InCombatLockdown() then
					rematch:SummonedPetMayChange()
					C_PetJournal.SetPetLoadOutInfo(loadoutSlot,petID)
				else
					-- "Pet battle already in progress."
					local info = ChatTypeInfo["SYSTEM"] -- use user's system chat color
					print(format("\124cff%02x%02x%02x%s",info.r*255,info.g*255,info.b*255,ERR_PETBATTLE_IN_BATTLE))
					print(REMATCH_IN_BATTLE)
					settings.CurrentLevelingPet = oldLevelingPet
					return
				end
			end
		end
	elseif not petID then
		settings.CurrentLevelingPet = nil
	end
	rematch:UpdateLevelingSlot()
	rematch:UpdateLevelingBorders()
	rematch:NoteLastLevelingPet()
end

-- if petID not in queue, adds to top of queue; if in queue, selects
function rematch:StartLevelingPet(petID)
	if not rematch:IsPetLeveling(petID) then
		table.insert(settings.LevelingQueue,1,petID)
		rematch:ProcessQueue(true) -- light processing to prevent redundantly finding a current leveling pet
	end
	rematch:SetCurrentLevelingPet(petID)
	if rematch.queue:IsVisible() then
		rematch:UpdateQueueList()
	end
end

-- adds or moves petID to the end of the queue
function rematch:AddLevelingPet(petID)
	local queue = settings.LevelingQueue
	if petID and rematch:PetCanLevel(petID) then
		if rematch:IsPetLeveling(petID) then
			for i=#queue,1,-1 do
				if queue[i]==petID then
					table.remove(queue,i)
				end
			end
		end
		table.insert(queue,petID)
		rematch:ProcessQueue()
	end
end

-- removes petID from the leveling queue
function rematch:StopLevelingPet(petID)
	if rematch:IsPetLeveling(petID) then
		local queue = settings.LevelingQueue
		for i=#queue,1,-1 do
			if queue[i]==petID then
				table.remove(queue,i)
			end
		end
		rematch:ProcessQueue()
	end
end

-- inserts petID into the index slot in the leveling queue
function rematch:InsertLevelingPet(petID,index)
	if petID and rematch:PetCanLevel(petID) then
		local queue = settings.LevelingQueue
		if queue[index]==petID then
			return -- pet already there!
		end
		if index>#queue then
			rematch:AddLevelingPet(petID) -- add to end of queue
			return
		end
		local queueSlot
		if rematch:IsPetLeveling(petID) then
			-- pet is in queue, we need to remove it from queue, but first find where it is in queue
			for i=1,#queue do
				if queue[i]==petID then
					queueSlot = i
					break
				end
			end
			if queueSlot<index then
				index = index - 1 -- and decrement index if it was in a slot we're about to remove
			end
			table.remove(queue,queueSlot) -- remove from queue
		end
		table.insert(queue,index,petID) -- insert into queue at index
		if index==1 then -- pets dragged to start also get set as current leveling pet
			rematch:SetCurrentLevelingPet(petID)
		end
		if queueSlot==1 then
			rematch:SetCurrentLevelingPet(queue[1])
		end
		rematch:ProcessQueue()
	end
end

-- this should be called whenever pets can possibly change
-- validates queue, sorts if needed and chooses new leveling pet if none after validation
function rematch:ProcessQueue(lightProcessing)
	if InCombatLockdown() then
		rematch.queueNeedsProcessed = true -- in case a pet levels to 25 when we're in combat
		return
	end
	if select(2,C_PetJournal.GetNumPets())==0 then -- if pets not loaded, come back in half a second to try again
		rematch:StartTimer("PetsRanAwayFromQueue",0.5,rematch.ProcessQueue)
		return
	end

	local queue = settings.LevelingQueue
	local hash = rematch.levelingPetsByID
	local sorted = rematch.sortedQueue

	-- rebuild levelingPetsByID and sortedQueue
	wipe(hash)
	wipe(sorted)
	for i=#queue,1,-1 do
		local petID = queue[i]
		local canLevel,level = rematch:PetCanLevel(petID)
		if canLevel and not hash[petID] then
			table.insert(sorted,1,petID)
			hash[petID] = i*100+level -- xxxyy.yy where xxx is raw queue index, yy.yy is pet level
		else
			table.remove(queue,i)
		end
	end

	-- sort the queue. order 1=ascending, 2=descending, 3=median
	if settings.QueueSortOrder then
		table.sort(sorted,function(e1,e2)
			local order = settings.QueueSortOrder
			local hash = rematch.levelingPetsByID
			local level1 = hash[e1]%100
			local level2 = hash[e2]%100
			if order==3 then -- for median sort, levels are distance from 10.5
				level1 = abs(level1-10.5)
				level2 = abs(level2-10.5)
			end
			if level1==level2 then -- if levels are the same, return sort by their original queue index
				return floor(hash[e1]/100)<floor(hash[e2]/100)
			else
				if order==2 then -- if descending, higher levels first
					return level1>level2
				else -- if ascending or median, lower levels (or distance from 10.5) first
					return level1<level2
				end
			end
		end)
	end

	-- post-processing stuff goes here; lightProcessing for calling functions that will do a proper process later
	if not lightProcessing then

		-- if Auto Rotate enabled, look for a new leveling pet if current leveling pet has gained xp
		if settings.QueueAutoRotate then
			local currentPetID = settings.CurrentLevelingPet
			if rematch.lastLevelingPetID==currentPetID and hash[currentPetID] then -- if leveling pet is unchanged and still leveling
				local newLevelingPetLevel = hash[currentPetID]%100
				if newLevelingPetLevel > rematch.lastLevelingPetLevel then -- and it's gained xp since last visit
					local order = settings.QueueSortOrder
					if order==1 or order==3 then -- ascending or median: get top-most pet
						if sorted[1] then
							rematch:SetCurrentLevelingPet(sorted[1])
						end
					elseif order==2 or not order then -- descending or unsorted: get next pet
						local nextPetID	= sorted[(rematch:GetCurrentQueuePosition() or 0)%#sorted+1]
						if nextPetID then
							rematch:SetCurrentLevelingPet(nextPetID)
						end
					end
				end
			end
		end

		-- if current leveling pet is no longer leveling, set a new one
		if not rematch:IsPetLeveling(settings.CurrentLevelingPet) then
			rematch:SetCurrentLevelingPet(queue[1])
		end

		-- update displayed list
		if rematch.queue:IsVisible() then
			rematch:UpdateQueueList()
		end

		rematch:NoteLastLevelingPet()

	end
end

-- called in both SetCurrentLevelingPet and ProcessQueue
-- notes the current leveling pet and its level for auto rotate purposes
function rematch:NoteLastLevelingPet()
	local petID = settings.CurrentLevelingPet
	if petID and rematch.levelingPetsByID[petID] then
		rematch.lastLevelingPetID = petID
		rematch.lastLevelingPetLevel = rematch.levelingPetsByID[petID]%100 or nil
	end
end

-- changes current leveling pet to the next or previous pet in the queue
function rematch:TraverseQueue(backward)
	rematch:DisablePrevNextButtons()
	if self.tooltipTitle then
		rematch.ShowTooltip(self) -- show tooltip again when buttons disabled
	end
	local dirOffset = backward and -2 or 0
	local index = ((rematch:GetCurrentQueuePosition() or 0)+dirOffset)%#rematch.sortedQueue+1
	local petID = rematch.sortedQueue[index]
	if petID then
		rematch:StartLevelingPet(petID)
	end
	if rematch.queue:IsVisible() then
		rematch:ScrollToQueueIndex(index)
	end
end

-- scrolls to an index in teamList and returns the button it scrolled to
function rematch:ScrollToQueueIndex(index)
	if index then
		local scrollFrame = rematch.queue.scrollFrame
		local buttons = scrollFrame.buttons
		local height = math.max(0,floor(scrollFrame.buttonHeight*(index-(#buttons)/2)))
		HybridScrollFrame_SetOffset(scrollFrame,height)
		scrollFrame.scrollBar:SetValue(height)
	end
end

function rematch:ScrollToQueuePetID(petID)
	if rematch.queue:IsVisible() then
		local queue = rematch.sortedQueue
		for i=1,#queue do
			if queue[i]==petID then
				rematch:ScrollToQueueIndex(i)
				return
			end
		end
	end
end

--[[ leveling switcher next/previous queue buttons ]]

-- when traversing through queue very fast, swaps can fall behind
-- need to disable the buttons momentarily when a new leveling pet needs to load
function rematch:DisablePrevNextButtons()
	local petID = settings.CurrentLevelingPet
	if C_PetJournal.PetIsSlotted(petID) then
		local slot
		for i=1,3 do
			if C_PetJournal.GetPetLoadOutInfo(i)==petID then
				slot = i
			end
		end
		rematch.queue.prevPet:Disable()
		rematch.queue.nextPet:Disable()
		rematch.levelingSwitcher.prevPet:Disable()
		rematch.levelingSwitcher.nextPet:Disable()
		rematch.queue.swapTimeout = 0
		rematch.queue.oldCurrentLevelingSlot = slot
		rematch:StartTimer("WaitForNewLevelingToLoad",0.25,rematch.EnablePrevNextButtons)
	end
end

function rematch:EnablePrevNextButtons()
	-- if new leveling pet is loaded (or we've exceeded a timeout after a re-attempt to load), enable buttons
	if C_PetJournal.PetIsSlotted(settings.CurrentLevelingPet) or rematch.queue.swapTimeout>6 then
		rematch.queue.prevPet:Enable()
		rematch.queue.nextPet:Enable()
		rematch.levelingSwitcher.prevPet:Enable()
		rematch.levelingSwitcher.nextPet:Enable()
		rematch.queue.swapTimeout = nil
		rematch.queue.oldCurrentLevelingSlot = nil
	else
		-- if new leveling pet isn't loaded yet, increment a timeout and come back again
		rematch.queue.swapTimeout = rematch.queue.swapTimeout + 1
		rematch:StartTimer("WaitForNewLevelingToLoad",0.25,rematch.EnablePrevNextButtons)
		if rematch.queue.swapTimeout>3 then -- if timeout>3 then try loading the pet one more time
			rematch:SummonedPetMayChange()
			C_PetJournal.SetPetLoadOutInfo(rematch.queue.oldCurrentLevelingSlot,settings.CurrentLevelingPet)
		end
	end
end

function rematch:LevelingSwitcherOnUpdate(elapsed)
	if not MouseIsOver(self) and not MouseIsOver(self:GetParent()) then
		self:Hide()
	end
end

function rematch:ShowLevelingSwitcher(parent)
	if parent.set==1 and parent.petID==rematch:GetCurrentLevelingPet() and rematch:GetNumLevelingPets()>1 then
	 -- onenter of leveling pet in current pets with more than one in queue
		rematch.levelingSwitcher:SetParent(parent)
		rematch.levelingSwitcher:SetPoint("TOP",parent,"BOTTOM",0,1)
		rematch.levelingSwitcher:SetFrameStrata("HIGH")
		rematch.levelingSwitcher:Show()
	end
end

--[[ leveling markers ]]

-- in ADDON_LOADED of the pet journal, HybridScrollFrame_Update is hooked to this
-- to go through all buttons in the journal to show/hide a leveling icon
function rematch:UpdateLevelingMarkersHook()
	if self==PetJournalListScrollFrame or self==PetJournalEnhancedListScrollFrame then
		for i=1,#self.buttons do
			local petID = self.buttons[i].petID
			local showIcon
			local icon = self.buttons[i].rematchLevelingPet
			if petID and rematch:IsPetLeveling(petID) then
				if not icon then
					self.buttons[i].rematchLevelingPet = self.buttons[i]:CreateTexture(nil,"ARTWORK")
					icon = self.buttons[i].rematchLevelingPet
					icon:SetSize(24,24)
					icon:SetPoint("RIGHT",-8,0)
					icon:SetTexture(rematch.levelingIcon)
				end
				showIcon = true
			end
			if icon then
				icon:SetShown(showIcon)
			end
		end
	end
end

-- call when pets are added or removed from leveling queue manually
function rematch:UpdateLevelingMarkers()
	if rematch.browser:IsVisible() then
		rematch:UpdateBrowser()
	elseif PetJournal and PetJournal:IsVisible() then
		if PetJournalEnhanced then
			rematch.UpdateLevelingMarkersHook(PetJournalEnhancedListScrollFrame)
		else
			rematch.UpdateLevelingMarkersHook(PetJournalListScrollFrame)
		end
	end
end

-- insertLine's OnUpdate that moves the line to where a battlepet on cursor will insert into queue
-- rematch.queue.scrollFrame.insertLine.index will be the index in the queue to insert the pet
-- an index of -1 means to add it to the end
function rematch:QueueListInsertLineOnUpdate(elapsed)
	self.timer = self.timer + elapsed
	if self.timer>0.1 then
		self.timer = 0
		local parent = self.parent
		if parent==rematch.queue.emptyCapture then
			self:SetPoint("CENTER",rematch.queue.emptyCapture,"TOP")
			self.index = -1 -- add to end of queue
		elseif parent then
			local _,y = GetCursorPosition()
			local top,bottom = parent:GetTop(),parent:GetBottom()
			local center = (top+bottom)*parent:GetEffectiveScale()/2
			if y>center then -- battlepet on cursor goes to slot previous to mouseover
				if top<=rematch.queue.scrollFrame:GetTop() then
					self:SetPoint("CENTER",parent,"TOP")
				else -- if above top of scrollFrame, get it off screen (but still "visible" so OnUpdate runs)
					self:ClearAllPoints()
				end
				self.index = parent.index
			else -- battlepet on cursor goes to slot after the mouseover
				if bottom>rematch.queue.scrollFrame:GetBottom()-6 then
					self:SetPoint("CENTER",parent,"BOTTOM")
				else -- move off screen if below bottom of scrollFrame
					self:ClearAllPoints()
				end
				self.index = parent.index+1
			end
		else
			self:Hide()
		end
	end
end

-- when pets in the queue receive a click/drag with a battlepet on the cursor
function rematch:QueueOnReceiveDrag()
	local cursorType,petID = GetCursorInfo()
	if cursorType=="battlepet" then
		if GetMouseButtonClicked()=="RightButton" then
			ClearCursor()
			rematch.queue.scrollFrame.insertLine:Hide()
			RematchTooltip:Hide()
			return
		elseif rematch:PetCanLevel(petID) then
			if settings.QueueSortOrder and rematch:IsPetLeveling(petID) then
				return -- can't reorder queue when there's a sort order
			elseif self==rematch.queue.emptyCapture or settings.QueueSortOrder then
				ClearCursor()
				rematch:AddLevelingPet(petID) -- emptyCapture click/drags always add to end of queue
				rematch.queue.scrollFrame.insertLine:Hide()
				RematchTooltip:Hide()
			else
				local index = rematch.queue.scrollFrame.insertLine.index
				if index then
					ClearCursor()
					rematch:InsertLevelingPet(petID,index) -- insert into the index derived by insertLine
					rematch.queue.scrollFrame.insertLine:Hide()
				end
			end
			rematch:UpdateLevelingMarkers()
		end
	end
end

function rematch:QueueOnDragStart()
	C_PetJournal.PickupPet(self.petID)
end

-- clicking of the sort+rotate control buttons
function rematch:QueueControlOnClick()
	local id = self:GetID()
	if id>0 then -- sort
		settings.QueueSortOrder = self:GetChecked() and id
		rematch:UpdateQueueControls()
		rematch:ProcessQueue()
		if not settings.KeepCurrentOnSort and rematch.sortedQueue[1] then
			rematch:SetCurrentLevelingPet(rematch.sortedQueue[1])
			rematch:UpdateQueueList()
		end
		rematch:ScrollToQueueIndex(1)
	else
		settings.QueueAutoRotate = self:GetChecked()
	end
end
